"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ts = require("typescript");
const path = require("path");
const fs_extra_p_1 = require("fs-extra-p");
const JsDocRenderer_1 = require("./JsDocRenderer");
const fs_extra_p_2 = require("fs-extra-p");
const util_1 = require("./util");
const __awaiter = require("./awaiter");
function generateAndWrite(basePath, config, tsConfig) {
    return __awaiter(this, void 0, void 0, function* () {
        let packageData = { name: "packageJsonNotDefined" };
        try {
            packageData = yield fs_extra_p_2.readJson(path.join(basePath, "package.json"));
        }
        catch (e) {
        }
        const generator = generate(basePath, config, tsConfig, packageData.name, packageData == null ? null : packageData.main);
        const out = path.resolve(basePath, tsConfig.jsdoc);
        console.log(`Generating JSDoc to ${out}`);
        yield fs_extra_p_1.emptyDir(out);
        for (const [moduleId, contents] of generator.moduleNameToResult.entries()) {
            yield fs_extra_p_1.writeFile(path.join(out, moduleId.replace(/\//g, "-") + ".js"), `/** 
 * @module ${moduleId}
 */

${contents.join("\n\n")}`);
        }
    });
}
exports.generateAndWrite = generateAndWrite;
function generate(basePath, config, tsConfig, moduleName, main) {
    const compilerOptions = config.options;
    const compilerHost = ts.createCompilerHost(compilerOptions);
    const program = ts.createProgram(config.fileNames, compilerOptions, compilerHost);
    util_1.checkErrors(ts.getPreEmitDiagnostics(program));
    const compilerOutDir = compilerOptions.outDir;
    if (compilerOutDir == null) {
        throw new Error("outDir is not specified in the compilerOptions");
    }
    const generator = new JsDocGenerator(program, path.relative(basePath, compilerOptions.outDir), moduleName, main, program.getCommonSourceDirectory());
    generateJsDoc(generator, path.resolve(basePath, tsConfig.jsdoc));
    return generator;
}
exports.generate = generate;
function generateJsDoc(generator, out) {
    for (const sourceFile of generator.program.getSourceFiles()) {
        if (!sourceFile.isDeclarationFile) {
            generator.generate(sourceFile);
        }
    }
}
class JsDocGenerator {
    constructor(program, relativeOutDir, moduleName, mainFile, commonSourceDirectory) {
        this.program = program;
        this.relativeOutDir = relativeOutDir;
        this.moduleName = moduleName;
        this.mainFile = mainFile;
        this.commonSourceDirectory = commonSourceDirectory;
        this.fileNameToModuleId = {};
        this.moduleNameToResult = new Map();
        this.renderer = new JsDocRenderer_1.JsDocRenderer(this);
    }
    sourceFileToModuleId(sourceFile) {
        if (sourceFile.isDeclarationFile) {
            if (sourceFile.fileName.endsWith("node.d.ts")) {
                return { sourceModuleId: "node", fileNameWithoutExt: "" };
            }
        }
        let sourceModuleId;
        const fileNameWithoutExt = sourceFile.fileName.slice(0, sourceFile.fileName.lastIndexOf(".")).replace(/\\/g, "/");
        const name = path.relative(this.commonSourceDirectory, fileNameWithoutExt);
        if (this.moduleName != null) {
            sourceModuleId = this.moduleName;
            if (name !== "index") {
                sourceModuleId += '/' + this.relativeOutDir;
            }
        }
        else {
            sourceModuleId = this.relativeOutDir;
        }
        if (name !== "index") {
            sourceModuleId += '/' + name;
        }
        if (this.mainFile == null ? fileNameWithoutExt.endsWith("/main") : `${fileNameWithoutExt}.js`.includes(path.posix.relative(this.relativeOutDir, this.mainFile))) {
            sourceModuleId = this.moduleName;
        }
        return { sourceModuleId, fileNameWithoutExt };
    }
    generate(sourceFile) {
        if (sourceFile.text.length === 0) {
            return;
        }
        this.currentSourceFile = sourceFile;
        this.renderer.currentSourceFile = sourceFile;
        const { sourceModuleId, fileNameWithoutExt } = this.sourceFileToModuleId(sourceFile);
        this.currentSourceModuleId = sourceModuleId;
        this.fileNameToModuleId[path.resolve(fileNameWithoutExt).replace(/\\/g, "/")] = sourceModuleId;
        const content = processTree(sourceFile, (node) => {
            if (node.kind === ts.SyntaxKind.InterfaceDeclaration || node.kind === ts.SyntaxKind.ClassDeclaration) {
                return this.renderClassOrInterface(node);
            }
            else if (node.kind === ts.SyntaxKind.FunctionDeclaration) {
                this.renderer.indent = "";
                return this.renderFunction(node);
            }
            else if (node.kind === ts.SyntaxKind.ExportKeyword) {
                return "";
            }
            else if (node.kind === ts.SyntaxKind.SourceFile) {
                return null;
            }
            else if (node.kind === ts.SyntaxKind.VariableStatement) {
                return this.renderVariableStatement(node);
            }
            return "";
        });
        let contents = this.moduleNameToResult.get(sourceModuleId);
        if (contents == null) {
            contents = [];
            this.moduleNameToResult.set(sourceModuleId, contents);
        }
        contents.push(content);
    }
    getTypeNamePathByNode(node) {
        if (node.kind === ts.SyntaxKind.UnionType) {
            const typeNames = [];
            for (const type of node.types) {
                const name = this.getTypeNamePathByNode(type);
                if (name == null) {
                    throw new Error("cannot get name for " + node.getText(node.getSourceFile()));
                }
                typeNames.push(name);
            }
            return typeNames.join(" | ");
        }
        else if (node.kind === ts.SyntaxKind.FunctionType) {
            return "callback";
        }
        else if (node.kind === ts.SyntaxKind.NumberKeyword) {
            return "number";
        }
        else if (node.kind === ts.SyntaxKind.StringKeyword) {
            return "string";
        }
        else if (node.kind === ts.SyntaxKind.BooleanKeyword) {
            return "boolean";
        }
        else if (node.kind === ts.SyntaxKind.NullKeyword) {
            return "null";
        }
        else if (node.kind === ts.SyntaxKind.UndefinedKeyword) {
            return "undefined";
        }
        else if (node.kind === ts.SyntaxKind.LiteralType) {
            const text = node.literal.text;
            return `"${text}"`;
        }
        const type = this.program.getTypeChecker().getTypeAtLocation(node);
        return type == null ? null : this.getTypeNamePath(type);
    }
    getTypeNamePath(type) {
        if (type.flags & ts.TypeFlags.Boolean) {
            return "boolean";
        }
        if (type.flags & ts.TypeFlags.Void) {
            return "void";
        }
        if (type.flags & ts.TypeFlags.Null) {
            return "null";
        }
        if (type.flags & ts.TypeFlags.String) {
            return "string";
        }
        if (type.flags & ts.TypeFlags.Number) {
            return "number";
        }
        if (type.flags & ts.TypeFlags.Undefined) {
            return "undefined";
        }
        const symbol = type.symbol;
        if (symbol == null || symbol.declarations == null || symbol.declarations.length === 0) {
            return null;
        }
        const valueDeclaration = symbol.valueDeclaration || ((symbol.declarations == null || symbol.declarations.length === 0) ? null : symbol.declarations[0]);
        if (ts.getCombinedModifierFlags(valueDeclaration) & ts.ModifierFlags.Ambient) {
            return symbol.name;
        }
        let typeSourceParent = valueDeclaration;
        while (typeSourceParent != null) {
            if (typeSourceParent.kind === ts.SyntaxKind.ModuleDeclaration && (typeSourceParent.flags & ts.NodeFlags.NestedNamespace) <= 0) {
                const m = typeSourceParent;
                const sourceModuleId = m.name.text;
                if (typeSourceParent.flags & ts.NodeFlags.Namespace) {
                    return `${sourceModuleId}:${symbol.name}`;
                }
                else {
                    return `module:${sourceModuleId}.${symbol.name}`;
                }
            }
            else if (typeSourceParent.kind === ts.SyntaxKind.SourceFile) {
                const sourceModuleId = this.sourceFileToModuleId(typeSourceParent).sourceModuleId;
                return `module:${sourceModuleId}.${symbol.name}`;
            }
            typeSourceParent = typeSourceParent.parent;
        }
        console.warn(`Cannot find parent for ${symbol}`);
        return null;
    }
    renderVariableStatement(node) {
        const flags = ts.getCombinedModifierFlags(node);
        if (!(flags & ts.ModifierFlags.Export)) {
            return "";
        }
        const declarations = node.declarationList == null ? null : node.declarationList.declarations;
        if (declarations == null && declarations.length !== 1) {
            return "";
        }
        const declaration = declarations[0];
        if (declaration.type == null) {
            return "";
        }
        this.renderer.indent = "";
        let typeName;
        const type = this.program.getTypeChecker().getTypeAtLocation(declaration);
        if (type.symbol != null && type.symbol.valueDeclaration != null) {
            typeName = this.getTypeNamePath(type);
        }
        else {
            typeName = declaration.type.typeName.text;
        }
        const tags = [`@type ${typeName}`];
        if (node.declarationList.flags & ts.NodeFlags.Const) {
            tags.push("@constant");
        }
        let result = this.renderer.formatComment(declaration, tags);
        result += `${this.renderer.indent}export var ${declaration.name.text}\n`;
        return result;
    }
    renderFunction(node) {
        const flags = ts.getCombinedModifierFlags(node);
        if (!(flags & ts.ModifierFlags.Export)) {
            return "";
        }
        const name = node.name.text;
        return this.renderer.renderMethod({ name: name, node: node, tags: [] });
    }
    renderClassOrInterface(node) {
        this.renderer.indent = "";
        const flags = ts.getCombinedModifierFlags(node);
        if (!(flags & ts.ModifierFlags.Export)) {
            return "";
        }
        const nodeDeclaration = node;
        const className = nodeDeclaration.name.text;
        let result = "/** \n";
        result += this.renderer.description(node);
        if (node.kind === ts.SyntaxKind.InterfaceDeclaration) {
            result += ` * @interface ${this.computeTypePath()}.${className}\n`;
        }
        const clazz = node;
        if (clazz.heritageClauses != null) {
            for (const heritageClause of clazz.heritageClauses) {
                if (heritageClause.types != null) {
                    for (const type of heritageClause.types) {
                        const typeNamePath = this.getTypeNamePathByNode(type);
                        if (typeNamePath != null) {
                            result += ` * @extends ${typeNamePath}\n`;
                        }
                    }
                }
            }
        }
        this.renderer.indent = "  ";
        const methods = [];
        const properties = [];
        for (const member of nodeDeclaration.members) {
            if (member.kind === ts.SyntaxKind.PropertySignature) {
                const p = this.describeProperty(member);
                if (p != null) {
                    properties.push(p);
                }
            }
            else if (member.kind === ts.SyntaxKind.MethodDeclaration || member.kind === ts.SyntaxKind.MethodSignature) {
                const m = this.renderMethod(member, className);
                if (m != null) {
                    methods.push(m);
                }
            }
        }
        result += this.renderer.renderProperties(properties);
        result += " */\n";
        result += `export class ${className} {\n`;
        methods.sort((a, b) => {
            let weightA = a.isProtected ? 100 : 0;
            let weightB = b.isProtected ? 100 : 0;
            weightA += trimMutatorPrefix(a.name).localeCompare(trimMutatorPrefix(b.name));
            return weightA - weightB;
        });
        for (const method of methods) {
            result += this.renderer.renderMethod(method);
            if (method !== methods[methods.length - 1]) {
                result += "\n";
            }
        }
        result += "}\n\n";
        return result;
    }
    describeProperty(node) {
        const flags = ts.getCombinedModifierFlags(node);
        if (flags & ts.ModifierFlags.Private) {
            return null;
        }
        const name = node.name.text;
        return { name, types: this.getTypeNamePathByNode(node.type), node };
    }
    renderMethod(node, className) {
        const flags = ts.getCombinedModifierFlags(node);
        if (flags & ts.ModifierFlags.Private) {
            return null;
        }
        const tags = [];
        const isProtected = (flags & ts.ModifierFlags.Protected) > 0;
        if (isProtected) {
            tags.push(`@protected`);
        }
        const name = node.name.text;
        tags.push(`@function ${this.computeTypePath()}.${className}#${name}`);
        return { name, tags, isProtected, node };
    }
    computeTypePath() {
        return "module:" + this.currentSourceModuleId;
    }
}
exports.JsDocGenerator = JsDocGenerator;
function processTree(sourceFile, replacer) {
    let code = '';
    let cursorPosition = 0;
    function skip(node) {
        cursorPosition = node.end;
    }
    function readThrough(node) {
        code += sourceFile.text.slice(cursorPosition, node.pos);
        cursorPosition = node.pos;
    }
    function visit(node) {
        readThrough(node);
        if (node.flags & ts.ModifierFlags.Private) {
            skip(node);
            return;
        }
        if (node.kind === ts.SyntaxKind.ImportDeclaration && node.importClause == null) {
            skip(node);
            return;
        }
        const replacement = replacer(node);
        if (replacement != null) {
            code += replacement;
            skip(node);
        }
        else {
            if (node.kind === ts.SyntaxKind.ClassDeclaration || node.kind === ts.SyntaxKind.InterfaceDeclaration || node.kind === ts.SyntaxKind.FunctionDeclaration) {
                code += "\n";
            }
            ts.forEachChild(node, visit);
        }
    }
    visit(sourceFile);
    code += sourceFile.text.slice(cursorPosition);
    return code;
}
exports.processTree = processTree;
function trimMutatorPrefix(name) {
    if (name.length > 4 && name[3] === name[3].toUpperCase() && (name.startsWith("get") || name.startsWith("set"))) {
        return name[3].toLowerCase() + name.substring(4);
    }
    return name;
}
//# sourceMappingURL=JsDocGenerator.js.map