#! /usr/bin/env node
"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ts = require("typescript");
const path = require("path");
const babel = require("babel-core");
const fs_extra_p_1 = require("fs-extra-p");
const bluebird_lst_1 = require("bluebird-lst");
const declarationGenerator_1 = require("./declarationGenerator");
const docGenerator_1 = require("./docGenerator");
const util_1 = require("./util");
const __awaiter = require("./awaiter");
util_1.transpile((basePath, config, tsConfig) => __awaiter(this, void 0, void 0, function* () {
    console.log(`Building ${basePath}`);
    const compilerOptions = config.options;
    let declarationConfig = tsConfig.declaration;
    if (declarationConfig === false) {
        declarationConfig = null;
    }
    if (declarationConfig != null) {
        compilerOptions.declaration = true;
    }
    compilerOptions.noEmitOnError = true;
    const program = ts.createProgram(config.fileNames, compilerOptions, ts.createCompilerHost(compilerOptions));
    util_1.checkErrors(ts.getPreEmitDiagnostics(program));
    const compilerOutDir = compilerOptions.outDir;
    if (compilerOutDir == null) {
        throw new Error("outDir is not specified in the compilerOptions");
    }
    yield fs_extra_p_1.ensureDir(compilerOutDir);
    const target = compilerOptions.target || ts.ScriptTarget.Latest;
    const fileToSourceMap = {};
    const promises = [];
    const emittedFiles = new Set();
    const declarationFiles = Array();
    const emitResult = program.emit(undefined, (fileName, data) => {
        emittedFiles.add(fileName);
        if (fileName.endsWith(".js")) {
            const sourceMapFileName = fileName + ".map";
            processCompiled(data, fileToSourceMap[sourceMapFileName], fileName, sourceMapFileName, promises);
        }
        else if (fileName.endsWith(".js.map")) {
            fileToSourceMap[fileName] = data;
        }
        else if (declarationConfig != null) {
            declarationFiles.push(ts.createSourceFile(fileName, data, target, true));
        }
    });
    util_1.checkErrors(emitResult.diagnostics);
    if (emitResult.emitSkipped) {
        throw new Error("Emit skipped");
    }
    if (declarationFiles.length > 0) {
        let packageData = null;
        try {
            packageData = yield fs_extra_p_1.readJson(path.join(basePath, "package.json"));
        }
        catch (e) {
        }
        const main = packageData == null ? null : packageData.main;
        if (typeof declarationConfig === "string") {
            declarationConfig = {
                [packageData.name]: declarationConfig,
            };
        }
        else if (declarationConfig === true) {
            declarationConfig = {
                [packageData.name]: path.join(path.resolve(compilerOutDir), `${packageData.name}.d.ts`),
            };
        }
        for (const moduleName of Object.keys(declarationConfig)) {
            promises.push(declarationGenerator_1.generateDeclarationFile(moduleName, declarationFiles, compilerOptions, path.resolve(basePath, declarationConfig[moduleName]), basePath, main));
        }
    }
    if (process.env.CI == null && tsConfig.docs != null) {
        const parsedDocs = docGenerator_1.generateDocs(program);
        const docs = docGenerator_1.renderDocs(parsedDocs);
        if (docs.length !== 0) {
            promises.push(docGenerator_1.writeDocFile(path.resolve(basePath, tsConfig.docs), docs));
        }
    }
    yield bluebird_lst_1.default.all(promises);
    yield removeOld(compilerOutDir, emittedFiles);
}))
    .catch(error => {
    console.error(error.stack || error.message || error);
    process.exit(-1);
});
function removeOld(outDir, emittedFiles) {
    return __awaiter(this, void 0, void 0, function* () {
        const files = yield fs_extra_p_1.readdir(outDir);
        yield bluebird_lst_1.default.map(files, file => {
            if (file[0] !== "." && !file.endsWith(".d.ts") && file !== "__snapshots__") {
                const fullPath = `${outDir}/${file}`;
                if (!file.includes(".")) {
                    return removeOld(fullPath, emittedFiles);
                }
                else if (!emittedFiles.has(fullPath)) {
                    return fs_extra_p_1.unlink(fullPath);
                }
            }
            return null;
        });
    });
}
function processCompiled(code, sourceMap, jsFileName, sourceMapFileName, promises) {
    const result = babel.transform(code, {
        inputSourceMap: sourceMap == null ? null : JSON.parse(sourceMap),
        sourceMaps: true,
        filename: jsFileName,
    });
    promises.push(fs_extra_p_1.outputFile(jsFileName, result.code), fs_extra_p_1.outputJson(sourceMapFileName, result.map));
}
//# sourceMappingURL=builder.js.map